import org.encog.Encog;
import org.encog.ml.data.MLDataSet;
import org.encog.ml.data.basic.BasicMLDataSet;
import org.encog.neural.networks.BasicNetwork;
import org.encog.neural.networks.layers.BasicLayer;
import org.encog.neural.networks.training.propagation.back.Backpropagation;
import org.encog.engine.network.activation.ActivationLOG;
import org.encog.util.csv.CSVFormat;
import org.encog.util.simple.EncogUtility;
import org.encog.util.simple.TrainingSetUtil;
import org.encog.mathutil.randomize.ConsistentRandomizer;
import java.io.*;
import java.util.Map;
import java.lang.String;

import org.encog.neural.networks.training.propagation.scg.ScaledConjugateGradient;
import org.encog.ml.train.strategy.HybridStrategy;
import org.encog.ml.train.MLTrain;

public class LearnThread extends Thread {
	Robot robot;
	boolean alive = true;
	private int maxDistance = 700;
    private int forwardSpeed = 5;
    private int turnSpeed = 3;
	private int poc = 0;
    private Writer out;
    private RobotHashMap<int[], int[]> map;
	
	LearnThread(Robot robot) {
		this.robot = robot;
		map = new RobotHashMap<int[], int[]>();
		map.setDifference(0);
		try {
        	out = new OutputStreamWriter(new FileOutputStream(new File("out.txt")));
    	} catch (Exception e){}
	}
	
	public void run() {
		while (alive) {
			learn();
		}
	}
	
	private boolean isNear(int dist) {
    	return dist > maxDistance;
    }
    
    protected boolean leftSideIsNear() {
    	return (isNear(robot.sensors[0].getDistValue()) || isNear(robot.sensors[1].getDistValue()) || isNear(robot.sensors[2].getDistValue()));
    }
    
    protected boolean rightSideIsNear() {
    	return (isNear(robot.sensors[3].getDistValue()) || isNear(robot.sensors[4].getDistValue()) || isNear(robot.sensors[5].getDistValue()));
    }
	
	public void learn() {
    	boolean left = leftSideIsNear(), right = rightSideIsNear();
    	
    	if (left) {
    		robot.turn(turnSpeed, -turnSpeed);
    		saveData(turnSpeed, -turnSpeed);
    		return;
    	}
    	if (right) {
    		robot.turn(-turnSpeed, turnSpeed);
    		saveData(-turnSpeed, turnSpeed);
    		return;
    	}
    	robot.forward();
    	saveData(forwardSpeed, forwardSpeed);
    }
    
    private void createNN() {
		robot.messagePasser.postMessage("learning... Please wait.");
		robot.messagePasser.postMessage("You can view process in command line.");
				
		BasicNetwork network = new BasicNetwork();
		MLDataSet trainingSet;
		int inputNeurons = 6, hiddenNeurons = 6, outputNeurons = 2, inputValue = 6, idealValue = 2;
		double nnError = 0.06;
		
		network.addLayer(new BasicLayer(new ActivationLOG(), true, inputNeurons));
		network.addLayer(new BasicLayer(new ActivationLOG(), true, hiddenNeurons));
		network.addLayer(new BasicLayer(new ActivationLOG(), false, outputNeurons));
		
		network.getStructure().finalizeStructure();
		network.reset();
		new ConsistentRandomizer(-1,1,500).randomize(network);
		
		trainingSet = TrainingSetUtil.loadCSVTOMemory(CSVFormat.ENGLISH, "out.txt", false, inputValue, idealValue);
		
		final Backpropagation train = new Backpropagation(network, trainingSet);
		train.fixFlatSpot(false);
				
		int epoch = 1;
 
		do {
			train.iteration();
			System.out.println("Epoch #" + epoch + " Error:" + train.getError());
			epoch++;
		} while(train.getError() > nnError);
				
		robot.setNN(network);
	}
	
	private void saveData(int leftSpeed, int rightSpeed) {
		int[] sensorsValues = new int[6], motorValues = new int[]{leftSpeed, rightSpeed};
		
		for (int i = 0; i < sensorsValues.length; i++) {
			sensorsValues[i] = robot.sensors[i].getDistValue();
		}
		
		if (!map.containsKey(sensorsValues)) {
			System.out.println(map.size());
			map.put(sensorsValues, motorValues);
			
			if (map.size() == 9000) {
				robot.stopLearn();
				writeToFile();
			}
		}
	}
    
    private void writeToFile() {  
		String line = "";
		try { 	
			for (Map.Entry<int[], int[]> entry : map.entrySet()) {
				for (int sensorsValue : entry.getKey()) {
					line += robot.normalizeSensorValue(sensorsValue)+",";
				}
				for (int mottorValue : entry.getValue()) {
					line += robot.normalizeSpeed(mottorValue)+",";
				}
				line = line.substring(0, line.length() - 1);
				out.write(line+"\n");
				line = "";
			}
			out.flush();
			out.close();
			
			createNN();
		} catch (Exception e) {}
    }
    
    public void stopLearn() {
		//~ map.clear();
		alive = false;
	}
	
	
}
