import org.encog.Encog;
import org.encog.neural.networks.BasicNetwork;

/*
 * Robot.java
 *
 * Created on July 27, 2001, 11:06 AM
 */

/**
 *
 * @author  Steve Perretta
 * @version
 *  
 */

import java.io.*;

public class Robot extends Thread {
   /* Changing the following fields is not recommended.
    * The variable types should NOT be altered, but if
    * you decide to change the variable names be sure to
    * change all occurences throughout the code. 
    */
    protected CurrentRobotState currentData;
    protected MessagePasser     messagePasser;
    protected Sensor[]          sensors; 
    protected Motor             motorState;
    private boolean           active;

    // Add your own fields here....
    private int forwardSpeed = 5;
    private int turnSpeed = 3;
    private boolean done = false;
    private LearnThread learnThread;
    private GoNNThread goNN;
    private BasicNetwork network;

   
    /** Constructor -- DO NOT MODIFY 
     */
    public Robot(CurrentRobotState data, MessagePasser ma) {
        currentData = data;
        messagePasser = ma;
        active = true;
    } 

    /** This method should be called from the run() method - before
      * entering the while loop. These assignments cannot be made in the
      * constructor (?)
      */
    private void getRobotObjects() {
        sensors  = currentData.getSensorValues();
        motorState = currentData.getMotorState();
    }   
    
    /** Essentially a "wrapper" that sets the foreward speed 
     */
    protected void forward() {
        motorState.setMotorSpeeds(forwardSpeed, forwardSpeed);
    }
            
    /** Another "wrapper" method for setting turns
     */
    protected void turn(int left, int right) {
        motorState.setMotorSpeeds(left, right);
    }
    
    /*  Terminates this thread. Called from the main program. 
     *    --- DO NOT REMOVE ---   
     */
    public void killControl() {
        active = false;
    }
       
    protected double normalizeSpeed(int speed) {
		return speed / (double) forwardSpeed;
	}
	
	protected int denormalizeSpeed(double speed) {
		return (int) Math.round(speed * forwardSpeed);
	}
	
	protected double normalizeSensorValue(int value) {
		double maxValue = 1023.0;
		
		return value / maxValue;
	}
	
	protected void setNN(BasicNetwork network) {
		messagePasser.postMessage("I know everything :). Press button 'g' to go.");
		messagePasser.postMessage("Press button 'g' to go.");
		this.network = network;
	}
	
	protected BasicNetwork getNN() {
		return network;
	}
	
	private void startLearn() {
		if (learnThread == null) {
			learnThread = new LearnThread(this);
			learnThread.start();
		}
	}
	
	protected void stopLearn() {
		if (learnThread != null) {
			learnThread.stopLearn();
			learnThread = null;
			motorState.setMotorSpeeds(0, 0);
		}
	}
	
	protected double[] normalizedSensorsValues() {
		double[] values = new double[6];
		
		for (int i = 0; i < values.length; i++) {
			values[i] = normalizeSensorValue(sensors[i].getDistValue());
		}
		
		return values;
	}
	
	private void NNGo() {
		if (goNN == null) {
			goNN = new GoNNThread(this);
			goNN.start();
		}
	}
	
	private void NNStop() {
		if (goNN != null) {
			goNN.stopNN();
			goNN = null;
			motorState.setMotorSpeeds(0, 0);
		}
	}
   
    private void execCommand(String command) {
		if (command.equals("l")) {
			messagePasser.postMessage("Getting training data...");
			startLearn();
			return;
		}
		if (command.equals("s")) {
			NNStop();
			return;
		}
		if (command.equals("g")) {
			messagePasser.postMessage("go");
			messagePasser.postMessage("For stoping me, press button 's'");
			NNGo();
			return;
		}
	}
        
    public void run() {
		String input = null;
		
        getRobotObjects();  // always call this !
        while(active) {
			input = messagePasser.getMessage();
			if (input != null) {
				execCommand(input);
			}
        }
    }
}


